"""
This module contains the following classes:
- RedisKey
- RedisList
- RedisHash
- RedisSet
- RedisString
- RedisDoubleHash
"""

from .redis_key import RedisKey
from .redis_list import RedisList
from .redis_hash import RedisHash
from .redis_set import RedisSet
from .redis_string import RedisString
from .redis_double_hash import RedisDoubleHash



"""
This module contains the following classes:
- RedisKey: represents a generic Redis key.
"""

from aioredis import Redis


class RedisKey:
    """
    Represents a Redis key of any type. Acts as the class for all data structures.
    """

    def __init__(self, redis: Redis, key: str):
        """
        Creates an instance ot `RedisKey`.

        Args:
            redis (Redis): The Redis instance to use to connect to Redis.
            key (str): The Redis key to use.
        """

        self._redis = redis
        self._key = key

    async def delete(self):
        """
        Deletes the key from Redis.
        """

        return await self._redis.delete(self._key)

    async def exists(self) -> bool:
        """
        Checks if the key exists in Redis or not.

        Returns:
            bool: A flag indicating whether the key exists or not.
        """

        return await self._redis.exists(self._key) > 0



from aioredis_models import RedisList
from .redis_tests import RedisTests

class RedisListTests(RedisTests):
    _key = 'test-key'
    _redis_list: RedisList = None

    async def asyncSetUp(self):
        await super().asyncSetUp()

        self._redis_list = RedisList(self._redis, self._key)
        await self._redis_list.delete()

    async def test_enumerate_gets_result(self):
        values = ['test', 'this', 'now', 'here', '!']
        await self._redis_list.push(*values, reverse=True)

        result = [item async for item in self._redis_list.enumerate(batch_size=2)]

        self.assertEqual(result, values)

    async def test_find_index_with_start_stop_batch_size_finds_result(self):
        values = ['this', 'is', 'a', 'very', 'valuable', 'test']
        await self._redis_list.push(*values, reverse=True)

        result = await self._redis_list.find_index('very', batch_size=2)

        self.assertEqual(result, values.index('very'))

    async def test_find_index_with_start_stop_batch_size_does_not_find_result_not_in_range(self):
        values = ['this', 'is', 'a', 'very', 'valuable', 'test']
        await self._redis_list.push(*values, reverse=True)

        result = await self._redis_list.find_index('valuable', start=1, stop=3, batch_size=2)

        self.assertIsNone(result)

    async def test_find_index_with_no_stop_finds_result(self):
        values = ['this', 'is', 'a', 'very', 'valuable', 'test']
        await self._redis_list.push(*values, reverse=True)

        result = await self._redis_list.find_index('valuable', start=1, batch_size=2)

        self.assertEqual(result, values.index('valuable'))

    async def test_find_index_with_no_stop_does_not_find_missing_result(self):
        values = ['this', 'is', 'a', 'very', 'valuable', 'test']
        await self._redis_list.push(*values, reverse=True)

        result = await self._redis_list.find_index('code', start=1, batch_size=2)

        self.assertIsNone(result)



from aioredis_models import RedisHash
from .redis_tests import RedisTests

class RedisHashTests(RedisTests):
    _key = 'test-key'
    _redis_hash: RedisHash = None

    async def asyncSetUp(self):
        await super().asyncSetUp()

        self._redis_hash = RedisHash(self._redis, self._key)
        await self._redis_hash.delete()

    async def test_enumerate_gets_result(self):
        values = {
            'foo': 'bar',
            'baz': 'bat',
            'boo': 'hoo',
            'snow': 'ball',
            'ball': 'crawl'
        }
        await self._redis_hash.set_all(values)

        result = {item async for item in self._redis_hash.enumerate(batch_size=2)}

        self.assertEqual(result, set(values.items()))



"""
This module contains the following classes:
- RedisSet: Represents a set stored in Redis.
"""

from typing import Set
from .redis_key import RedisKey


class RedisSet(RedisKey):
    """
    Represents a set stored in Redis.
    """

    async def size(self) -> int:
        """
        Gets the size of the set.

        Returns:
            int: The size of the set.
        """

        return await self._redis.scard(self._key)

    async def get_all(self, encoding='utf-8') -> Set:
        """
        Gets all the members of the set.

        Args:
            encoding (str, optional): The encoding to use when decoding set members. Defaults to
                'utf-8'.

        Returns:
            Set: The members of the set.
        """

        return await self._redis.smembers(self._key, encoding=encoding)

    async def add(self, value: str) -> int:
        """
        Adds an item to the set.

        Args:
            value (str): The item to add.

        Returns:
            int: The number of items that were added to the set.
        """

        if value is not None:
            return await self._redis.sadd(self._key, value)

    async def remove(self, value: str) -> int:
        """
        Removes an item from the set.

        Args:
            value (str): The item to remove.

        Returns:
            int: The number of elements that were removed from the set.
        """

        return await self._redis.srem(self._key, value)



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
setup(
    name="aioredis-models",
    version="2.0.0",
    description="Model Redis data structures",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/qcrisw/aioredis-models",
    author="QCRI Software Group",
    author_email="qcriswgroup@hbku.edu.qa",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords = ['redis', 'asyncio', 'data-structures', 'models'],
    packages=find_packages(exclude=("tests",)),
    install_requires=["aioredis==1.3.1"],
)



"""
This module contains the following classes:
- RedisDoubleHash: Represents a two-way hash map stored in Redis.
"""

from typing import Set
# Aliasing this to avoid confusion with the `set` function below.
from builtins import set as builtin_set
from aioredis import Redis
from .redis_key import RedisKey
from .redis_set import RedisSet


class RedisDoubleHash:
    """
    Represents a two-way hash map stored in Redis. Each field can be associated with multiple
    values and each value can be associated with multiple fields. The structure allows for
    getting all the values for a field and all the fields for a given value. The values are
    thus referred to as inverted fields.
    """

    def __init__(self, redis: Redis, key: str, inverse_key: str):
        """
        Creates an instance of `RedisDoubleHash`.

        Args:
            redis (Redis): The Redis instance to use to connect to Redis.
            key (str): The key to use for forward hash map.
            inverse_key (str): The key to use for inverted hash map.
        """

        self._redis = redis
        self._key = key
        self._inverse_key = inverse_key

    async def fields(self) -> Set:
        """
        Gets all the fields in the hash map.

        Returns:
            Set: The fields in the hash map.
        """

        return builtin_set(
            self._extract_field_name(redis_key, self._key) \
                for redis_key in await self._fields_generic()
        )

    async def fields_inverted(self) -> Set:
        """
        Gets all the inverted fields (values) in the hash map.

        Returns:
            Set: The inverted fields (values) in the hash map.
        """

        return builtin_set(
            self._extract_field_name(redis_key, self._inverse_key) \
                for redis_key in await self._fields_generic(inverse=True)
        )

    async def get(self, field: str) -> Set:
        """
        Gets the inverted fields associated with the given field.

        Args:
            field (str): The field to get.

        Returns:
            Set: The set of all inverted fields associated with the given field.
        """

        return await self._get_field_value(self._key, field)

    async def get_inverted(self, field: str) -> Set:
        """
        Gets the fields associated with the given inverted field (value).

        Args:
            field (str): The inverted field to get.

        Returns:
            Set: The set of all fields associated with the given inverted field.
        """

        return await self._get_field_value(self._inverse_key, field)

    async def set(self, field: str, value: str):
        """
        Associates the given value with the given field.

        Args:
            field (str): The name of the field.
            value (str): The value to associate.
        """

        if value is None:
            return
        await self._set_field(self._key, field, value)
        await self._set_field(self._inverse_key, value, field)

    async def unset(self, field: str, value: str):
        """
        Dissociates the given value with the given field.

        Args:
            field (str): The name of the field.
            value (str): The name of the value to dissociate.
        """

        if value is None:
            return
        await self._unset_field(self._key, field, value)
        await self._unset_field(self._inverse_key, value, field)

    async def set_inverted(self, field: str, value: str):
        """
        Associates the given value with the inverted field.

        Args:
            field (str): The name of the inverted field.
            value (str): The value to associate.
        """

        if value is None:
            return
        return await self.set(field=value, value=field)

    async def remove(self, field: str):
        """
        Removes the given field from both sides of the hash map.

        Args:
            field (str): The field to remove.
        """

        return await self._remove_generic(self._key, self._inverse_key, field)

    async def remove_inverted(self, field: str):
        """
        Removes the given inverted field from both sides of the hash map.

        Args:
            field (str): The inverted field to remove.
        """

        return await self._remove_generic(self._inverse_key, self._key, field)

    async def delete(self):
        """
        Deletes all mappings from both sides of the hash map.
        """

        await self._sub_delete(False)
        await self._sub_delete(True)

    async def _fields_generic(self, inverse: bool=False) -> Set:
        return await self._redis.keys(self._get_field_name(
            self._inverse_key if inverse else self._key,
            '*'
        ), encoding='utf-8')

    async def _get_field_value(self, key: str, field: str) -> Set:
        sub_set = self._get_redis_set(key, field)
        return await sub_set.get_all()

    async def _set_field(self, key: str, field: str, value: str):
        sub_set = self._get_redis_set(key, field)
        return await sub_set.add(value)

    async def _unset_field(self, key: str, field: str, value: str):
        sub_set = self._get_redis_set(key, field)
        return await sub_set.remove(value)

    async def _sub_delete(self, inverse: bool):
        for key in await self._fields_generic(inverse):
            await RedisKey(self._redis, key).delete()

    async def _remove_generic(self, key: str, inverse_key: str, field: str):
        sub_set = self._get_redis_set(key, field)
        values = await sub_set.get_all()
        await sub_set.delete()
        for value in values:
            value_set = RedisSet(self._redis, self._get_field_name(inverse_key, value))
            await value_set.remove(field)

    def _get_redis_set(self, key: str, field: str) -> RedisSet:
        return RedisSet(self._redis, self._get_field_name(key, field))

    @staticmethod
    def _get_field_name(key: str, field: str) -> str:
        return f'{key}:{field}'

    @staticmethod
    def _extract_field_name(redis_key: str, key: str) -> str:
        return redis_key.removeprefix(key + ':')
